import asyncio
import json
import random
import time
import uuid
from asyncio import StreamReader, StreamWriter
from dataclasses import dataclass, asdict

from tanks.common import Tank, Direction, GAME_AREA_WIDTH, GAME_AREA_HEIGHT, Missile


@dataclass
class ConnectedClient:
    identifier: str
    reader_stream: StreamReader
    writer_stream: StreamWriter


class Server:
    def __init__(self) -> None:
        self.connected_clients: list[ConnectedClient] = []
        self.tanks: dict[str, Tank] = {}
        self.missiles: dict[str, Missile] = {}

    async def handle_new_client(
        self, reader_stream: StreamReader, writer_stream: StreamWriter
    ) -> None:
        new_client = self.setup_new_client(reader_stream, writer_stream)
        await self.notify_clients_about_gameplay()
        while True:
            client_message = await self.read_user_data(reader_stream)
            self.recalculate_position_gameplay(client_message, new_client.identifier)
            await self.notify_clients_about_gameplay()
            if client_message == "q":
                await self.user_quit(new_client)
                return

    async def read_user_data(self, reader_stream):
        user_data = await reader_stream.read(100)
        client_message = user_data.decode()
        return client_message

    async def user_quit(self, new_client):
        del self.tanks[new_client.identifier]
        self.connected_clients.remove(new_client)
        await self.notify_clients_about_gameplay()

    def setup_new_client(
        self, reader_stream: StreamReader, writer_stream: StreamWriter
    ) -> ConnectedClient:
        client_id = str(uuid.uuid1())
        new_client = ConnectedClient(
            identifier=client_id, reader_stream=reader_stream, writer_stream=writer_stream
        )
        self.connected_clients.append(new_client)
        print(f"Dołączył nowy klient - id: {new_client.identifier}")
        self.setup_new_tank(new_client)
        return new_client

    def setup_new_tank(self, new_client) -> None:
        new_tank_y = random.randint(10, 40)
        new_tank_x = random.randint(10, 110)
        new_tank = Tank(center_y=new_tank_y, center_x=new_tank_x, direction=Direction.UP)
        self.tanks[new_client.identifier] = new_tank
        print(
            f"Nowy czołg --> Y: {new_tank.center_y} X: {new_tank.center_x} D: {new_tank.direction}"
        )

    async def notify_clients_about_gameplay(self) -> None:
        tanks_data = [asdict(tank) for tank in self.tanks.values()]
        missiles_data = [asdict(missile) for missile in self.missiles.values()]
        message_data = {"tanks": tanks_data, "missiles": missiles_data}
        message = json.dumps(message_data).encode()
        for client in self.connected_clients:
            client.writer_stream.write(message)
            await client.writer_stream.drain()

    def recalculate_position_gameplay(self, user_input: str, client_id: str) -> None:
        if user_input == "w":
            self.tanks[client_id].center_y -= 1
            self.tanks[client_id].direction = Direction.UP
        elif user_input == "s":
            self.tanks[client_id].center_y += 1
            self.tanks[client_id].direction = Direction.DOWN
        elif user_input == "a":
            self.tanks[client_id].center_x -= 2
            self.tanks[client_id].direction = Direction.LEFT
        elif user_input == "d":
            self.tanks[client_id].center_x += 2
            self.tanks[client_id].direction = Direction.RIGHT
        elif user_input == "m" and client_id not in self.missiles:
            new_missile = Missile(
                center_x=self.tanks[client_id].center_x,
                center_y=self.tanks[client_id].center_y,
                direction=self.tanks[client_id].direction,
            )
            self.missiles[client_id] = new_missile

        if self.tanks[client_id].center_x > GAME_AREA_WIDTH - 3:
            self.tanks[client_id].center_x = GAME_AREA_WIDTH - 3
        elif self.tanks[client_id].center_x < 3:
            self.tanks[client_id].center_x = 3
        if self.tanks[client_id].center_y > GAME_AREA_HEIGHT - 2:
            self.tanks[client_id].center_y = GAME_AREA_HEIGHT - 2
        elif self.tanks[client_id].center_y < 2:
            self.tanks[client_id].center_y = 2

    async def handle_missiles(self) -> None:
        while True:
            await asyncio.sleep(0.1)
            for client_id, missile in self.missiles.items():
                if missile.direction == Direction.UP:
                    missile.center_y -= 1
                elif missile.direction == Direction.DOWN:
                    missile.center_y += 1
                elif missile.direction == Direction.LEFT:
                    missile.center_x -= 2
                elif missile.direction == Direction.RIGHT:
                    missile.center_x += 2

            self.missiles = {
                client_id: missile
                for client_id, missile in self.missiles.items()
                if missile.is_in_game_area()
            }
            await self.notify_clients_about_gameplay()


async def run_server() -> None:
    server_handler = Server()
    loop = asyncio.get_event_loop()
    loop.create_task(server_handler.handle_missiles())
    server = await asyncio.start_server(server_handler.handle_new_client, "127.0.0.1", 8888)
    async with server:
        await server.serve_forever()
